#! /usr/bin/python3

import numpy as np
from scipy import signal
from skimage.measure import block_reduce

def round_to( x, bits_prec ):
    factor = 1 << bits_prec
    return np.round( x * factor )/factor

def floor_to( x, bits_prec ):
    factor = 1 << bits_prec
    return np.floor( x * factor )/factor

def get_ternary( x ):
    '''
    Useful for networks such as ternary weight networks where
    the values of the network is -a,0,a
    This function returns the ( ( weights as -1,0,1 ), a )
    '''
    x_flat = x.flatten()
    scaling_factor = np.mean(abs(x_flat[abs(x_flat) > 0]))
    tri_weights = np.round(x/scaling_factor).astype( int )
    return tri_weights, scaling_factor

def conv2d( img, conv_weights, padding = 'same' ):
    '''
    img => img to convolve
    conv_weights => the weights for the convolution
    padding => "same" or "valid"
    '''
    filter_out = []
    for fno in range( conv_weights.shape[-1] ):
        ch_sum = 0
        for chno in range( conv_weights.shape[-2] ):
            conv_window = np.flip( np.flip( conv_weights[:,:,chno,fno], 0 ), 1 )
            ch_sum += signal.convolve2d( img[:,:,chno], conv_window, mode = padding )
        filter_out += [ ch_sum ]
    filter_out = np.array( filter_out )
    return np.transpose( filter_out, [ 1, 2, 0 ] )

def conv1d( img, conv_weights, padding = "same" ):
    '''
    img => img to convolve
    conv_weights => the weights for the convolution
    padding => "same" or "valid"
    '''
    filter_out = []
    for fno in range( conv_weights.shape[-1] ):
        ch_sum = 0
        for chno in range( conv_weights.shape[-2] ):
            conv_window = np.flip( conv_weights[:,chno,fno], 0 )
            ch_sum += signal.convolve( img[:,chno], conv_window, mode = padding )
        filter_out += [ ch_sum ]
    filter_out = np.array( filter_out )
    return np.transpose( filter_out, [ 1, 0 ] )

def get_AB( mean, var, gamma, beta, scaling_factor, bias = 0, thres = 10**15 ):
    '''
    Merge BN and TWN variables into a transform: ax + b
    returns ( a, b )
    '''
    mean = mean - bias
    stddev = np.sqrt( var )
    a = gamma / stddev
    a = a * ( a < thres ) # if very big then prob is ignored anyway: ie) stddev = 0
    b = beta - a * mean
    a = a * scaling_factor
    return ( a, b )

def get_AB_quantized( mean, var, gamma, beta, s_in, bits, is_round = True ):
    '''
    Get variables to perform batch norm with a limited number of bits in and out
    mean, var, gamma, beta => floating point variables from batch norm
    s_in => the amount to input is scaled by
    bits => the number of bits the output should be
    is_round => if a rounding function was used to quantize or if False, assumes ceil function
    returns ( a, b, x_min, x_max )
    if x < x_min => 0
    if x >= x_max => ( 1 << bits ) - 1
    else use ( a*x + b ) >> ( bits + 2 )
    '''
    a = gamma / np.sqrt( var + 0.001 )
    b = beta - a * mean
    s_out = ( 1 << bits ) - 1
    a *= s_out/s_in
    b *= s_out
    if is_round:
        x_min = np.floor( s_in*( 1/2 - b )/a)
        x_max = np.ceil( s_in*(s_out - 1/2 - b )/a)
        b += 0.4999999
    else:
        x_min = np.floor(-b*s_in/a)
        x_max = np.ceil( s_in*( s_out - b )/a )
        b += 0.999999
    prec_s = ( 1 << ( bits + 2 ) )
    a = np.round( a * prec_s ) / prec_s
    b = np.round( b * prec_s ) / prec_s
    return a, b, x_min, x_max

def maxpool2d( img, kern_and_stride = 2 ):
    return block_reduce( img, (kern_and_stride, kern_and_stride, 1), np.max )

def maxpool1d( img, kern_and_stride = 2 ):
    return block_reduce( img, (kern_and_stride, 1), np.max )

def relu( img ):
    return img * ( img > 0 )



#! /usr/bin/python3

import csv

def make_ROM( weights, no_vecs, bw_w, rom_name = "w_mem" ):
    assert len(weights) % no_vecs == 0, "len(weights) must be divisible by no_vecs"
    fmt = "00" + str( bw_w ) + "b"
    no_cycs = int( len(weights) / no_vecs )
    no_bits_each_cyc = no_vecs*bw_w * len( weights[0] )
    rom_txt = "reg [" + str( no_cycs - 1 ) + ":0][" + str( len(weights[0]) - 1 ) + ":0][" + str( no_vecs*bw_w - 1) + ":0]" + rom_name + ";\n"
    rom_txt += "initial begin\n"
    for i in range( no_cycs ):
        rom_txt += rom_name + "[" + str(i) + "] = "
        rom_txt += str( no_bits_each_cyc ) + "'b"
        subset = np.transpose( weights[i*no_vecs:(i+1)*no_vecs] )
        for j in range( len( subset ) ):
            rom_txt += "".join( [ format( v + ( 1 << bw_w ), fmt )[-bw_w:] for v in subset[j] ] )
        rom_txt += ";\n"
    rom_txt += "end\n"
    return rom_txt

def DMM_generate( fname_w, module_name, BW_in, BW_out, cycles ):
    f = open( fname_w )
    rdr = csv.reader( f )
    weights = np.array( [ [ int(x) for x in y ] for y in rdr ] )
    f.close()
    max_val = np.max( weights )
    if max_val > 0:
        max_val += 1  # eg) log2(2) => 1 but needs 2 bits
    bits_max = np.ceil( np.log2( np.abs( max_val ) ))
    min_val = np.min( weights )
    if min_val != 0:
        bits_min = np.ceil( np.log2( np.abs( min_val ) ))
    else:
        bits_min = 0
    # determine the bitwidth needed for the weights
    bits_w = max( bits_max, bits_min )
    if min_val < 0: # add a bit for sign
        bits_w += 1
    no_inputs = len(weights)
    no_outputs = len( weights[0] )
    no_vecs = int( no_inputs/cycles )
    log2_cyc = int( np.ceil( np.log2( cycles ) ) )
    log2_no_vecs = int( np.ceil( np.log2( no_vecs ) ) )
    module_header = "module " + module_name + """ (
input clock,
input reset,
input vld_in,
input [""" + str( no_vecs - 1) + ":0][" + str(BW_in-1) + """:0] in,
output vld_out,
output [""" + str( no_outputs - 1) + ":0][" + str(BW_out-1) + """:0] out
);
"""
    module_body = make_ROM( weights, no_vecs, bits_w, "w_mem" )
    if cycles > 1:
        module_body += "reg [" + str( log2_cyc - 1 ) + """:0] cntr;
always @( posedge clock ) begin
  if ( reset ) begin
    cntr <= 0;
  end else begin
    if ( vld_in ) begin
      if ( cntr >= """ + str( cycles - 1 ) + """) begin
         cntr <= 0;
      end else begin
         cntr <= cntr + 1;
      end
    end
  end
end
"""
    else: # if all in one cycle
        module_body += "wire cntr;\nassign cntr = 0;\n"
    # w_mem[cntr][i] gets the weights for this cycle and filter i
    module_body += """
genvar i;
generate
multiply_accumulate
  #(
    .LOG2_NO_VECS(""" + str(log2_no_vecs) + """),
    .BW_IN(""" + str( BW_in ) + """),
    .BW_OUT(""" + str( BW_out ) + """),
    .BW_W(""" + str( bits_w ) + """)
) (
  .clk(clock),
  .new_sum(cntr == 0),
  .data_in(in),
  .w_vec( w_mem[cntr] ),
  .data_out( out[i] )
);
endgenerate
    """
    # make the ROM with the weights values
    module_body += "endmodule\n"
    return module_header + module_body



#! /usr/bin/python3

import csv
from . import add_ops

def get_name( idx, no_inputs ):
    if idx < 0: # if unused then is 0
        return "0"
    if idx < no_inputs:
        return "in[" + str(idx) + "]"
    return "tree_" + str(idx - no_inputs)

def get_max_d( depths ):
    return max( [ depths[i] for i in depths ] )

def create_vld_and_reset( max_depth ):
    assert max_depth >= 1, "Must have atleast 1 stage"
    module_vld_rst = "reg [" + str( max_depth - 1 ) + ":0] vld_reg = 0;\n"
    module_vld_rst += "wire [" + str( max_depth - 1 ) + ":0] resets;\n"
    if max_depth >= 2:
        module_vld_rst += "reg [" + str( max_depth - 2 ) + ":0] rst_reg;\n"
        module_vld_rst += "assign resets = { rst_reg, reset };\n"
    else:
        module_vld_rst += "assign resets[0] = reset;\n"
    module_vld_rst += "assign vld_out = vld_reg[" + str(max_depth - 1) + "];\n"
    module_vld_rst += "always @( posedge clock ) begin\n"
    if max_depth >= 2:
        module_vld_rst += "vld_reg <= { vld_reg[" + str( max_depth - 2) + ":0], vld_in };\n"
    else:
        module_vld_rst += "vld_reg[0] <= vld_in;\n"
    if max_depth >= 2:
        module_vld_rst += "rst_reg <= resets[" + str(max_depth-2) + ":0];\n"
    module_vld_rst += "end\n"
    return module_vld_rst

def set_outputs( depths, output_idxs, bws, BW, no_inputs, signed_ext = True ):
    # make sure all the outputs match the correct depth
    max_depth = get_max_d( depths )
    declarations = []
    computations = []
    for idx, oi in enumerate( output_idxs ):
        delay_reg = max_depth - depths[oi]
        padded_out = get_name( oi, no_inputs )
        if signed_ext and bws[oi] < BW:
            padded_out = "{ {" + str( BW-bws[oi] ) + "{" + padded_out + "[" + str( bws[oi]-1 ) + "]}}," + padded_out + "}"
        if delay_reg == 0 or get_name( oi, no_inputs ) == '0':
            declarations += [ "assign out[" + str(idx) + "] = " + padded_out + ";" ]
        elif delay_reg == 1:
            declarations += [ "reg [" + str(BW-1) + ":0] out_" + str(idx) + ";" ]
            declarations += [ "assign out[" + str(idx) + "] = out_" + str(idx) + ";" ]
            computations += [ "out_" + str(idx) + " <= " + padded_out + ";" ]
        else:
            declarations += [ "reg [" + str(delay_reg-1) + ":0][" + str(BW-1) + ":0] out_" + str(idx) + ";" ]
            declarations += [ "assign out[" + str(idx) + "] = out_" + str(idx) + "[" + str(delay_reg - 1) + "];" ]
            computations += [ "out_" + str(idx) + " <= { out_" + str(idx) + "[" + str( delay_reg - 2 ) + ":0], " + padded_out + "};" ]
    output_txt = "\n".join( declarations )
    output_txt += "\nalways @( posedge clock ) begin\n"
    output_txt += "\n".join( computations )
    output_txt += "\nend\n"
    return output_txt

def SMM_generate( fname_ops, module_name, BW_in, BW_out, create_op ):
    f = open( fname_ops )
    rdr = csv.reader( f )
    ops = [ [ int(x) for x in y ] for y in rdr ]
    output_idxs = ops[0]
    ops = ops[1:]
    no_inputs = ops[0][0]
    module_header = "module " + module_name + """ (
input clock,
input reset,
input vld_in,
input [""" + str( no_inputs - 1) + ":0][" + str(BW_in-1) + """:0] in,
output vld_out,
output [""" + str( len(output_idxs) - 1) + ":0][" + str(BW_out-1) + """:0] out
);
"""
    depths = {}
    bws = {}
    depths[-1] = 0
    for i in range( no_inputs ):
        depths[i] = 0
        bws[i] = BW_in
    module_body = ""
    declarations = []
    computations = []
    # implement all the ops
    for op in ops:
        stage_reset = "resets[" + str(depths[op[1]]) + "]"
        names = [ get_name( idx, no_inputs ) for idx in op[:4] ]
        module_txt, op_BW = create_op( names, op[4], op[5:], BW_in, BW_out, "op_" + str(op[0]), stage_reset, depths[op[1]] )
        module_body += module_txt
        bws[op[0]] = op_BW
        if op[2] != -1:
            assert depths[op[1]] == depths[op[2]], "Depths mismatch, invalid tree in op = " + str( op )
        depths[op[0]] = depths[op[1]] + 1
    max_depth = get_max_d( depths )
    module_header += create_vld_and_reset( max_depth )
    module_body += set_outputs( depths, output_idxs, bws, BW_out, no_inputs )
    module_body += "endmodule\n"
    return module_header + module_body
